@@ -7,6 +7,7 @@ Numba version of bias-corrected and accelerated (BCa) bootstrap.
 import numpy as np
 import pandas as pd
 from numba import njit, boolean
+from numba.core.registry import CPUDispatcher
 from itertools import product
 from collections import defaultdict
 from collections.abc import Iterable
@@ -52,7 +53,7 @@ def resample_paired_nb(X, Y, func, output_len=1, R=int(1e5), seed=0):
 
 @njit
 def resample_nb_X(X, R=int(1e5), seed=0, smooth=False, N=0):
-    """X: array of shape (N_samples, n_vars)."""
+    """X: array of shape (N, n_vars)."""
     np.random.seed(seed)
     n, num_vars = X.shape
     if N == 0:
@@ -73,10 +74,25 @@ def resample_nb_X(X, R=int(1e5), seed=0, smooth=False, N=0):
             data_resampled[:,:,k] += h_k * np.random.standard_t(n_trimmed, R*N)
     return data_resampled
 
+@njit
+def resample_nb_X_multidim(X, R=int(1e5), seed=0, N=0):
+    """X: array of shape (N, *dims)."""
+    np.random.seed(seed)
+    n = X.shape[0]
+    if N == 0:
+        N = n
+    idxs_resampling = np.random.randint(low=0, high=n, size=R*N)
+    data_resampled = X[idxs_resampling].reshape(R, N, *X.shape[1:])
+    return data_resampled
+
 @njit
 def resample_nb(X, func, output_len=1, R=int(1e5), seed=0, smooth=False, N=0):
-    """X: array of shape (N_samples, n_vars)."""
-    data_resampled = resample_nb_X(X, R=R-1, seed=seed, smooth=smooth, N=N)
+    """X: array of shape (N, n_vars)."""
+    if X.ndim > 2:
+        assert not smooth, "Smooth not supported for multidimensional data."
+        data_resampled = resample_nb_X_multidim(X, R=R-1, seed=seed, N=N)
+    else:
+        data_resampled = resample_nb_X(X, R=R-1, seed=seed, smooth=smooth, N=N)
 
     boot_sample = np.empty((R, output_len))
     boot_sample[0] = func(X) # first element is the sample statistic.
@@ -86,7 +102,7 @@ def resample_nb(X, func, output_len=1, R=int(1e5), seed=0, smooth=False, N=0):
 
 @njit
 def resample_twosamples_nb(X1, X2, func, output_len=1, R=int(1e5), seed=0, smooth=False, N=0):
-    """Xi: array of shape (N_samples, n_vars)."""
+    """Xi: array of shape (N, n_vars)."""
     data_resampled_1 = resample_nb_X(X1, R=R-1, seed=seed, smooth=smooth, N=N)
     data_resampled_2 = resample_nb_X(X2, R=R-1, seed=seed+1, smooth=smooth, N=N)
 
@@ -97,7 +113,7 @@ def resample_twosamples_nb(X1, X2, func, output_len=1, R=int(1e5), seed=0, smoot
     return boot_sample
 
 def resample_twosamples(X1, X2, func, output_len=1, R=int(1e5), seed=0, smooth=False, N=0):
-    """Xi: array of shape (N_samples, n_vars)."""
+    """Xi: array of shape (N, n_vars)."""
     data_resampled_1 = resample_nb_X(X1, R=R-1, seed=seed, smooth=smooth, N=N)
     data_resampled_2 = resample_nb_X(X2, R=R-1, seed=seed, smooth=smooth, N=N)
 
@@ -195,8 +211,12 @@ def resample_block_nb(X, Y, func, output_len=1, R=int(1e4), R_B=int(1e3), seed=0
 #     return boot_sample
 
 def resample(X, func, output_len=1, R=int(1e4), seed=0, smooth=False, N=0):
-    """X: array of shape (N_samples, n_vars)."""
-    data_resampled = resample_nb_X(X, R=R-1, seed=seed, smooth=smooth, N=N)
+    """X: array of shape (N, *dims)."""
+    if X.ndim > 2:
+        assert not smooth, "Multidimensional resampling does not support smoothing."
+        data_resampled = resample_nb_X_multidim(X, R=R-1, seed=seed, N=N)
+    else:
+        data_resampled = resample_nb_X(X, R=R-1, seed=seed, smooth=smooth, N=N)
 
     boot_sample = np.empty((R, output_len))
     boot_sample[0] = func(X) # first element is the sample statistic.
@@ -268,6 +288,11 @@ def jackknife_stat_nb(data, statistic):
     stats = np.array([statistic(r) for r in resamples])
     return stats
 
+def jackknife_stat(data, statistic):
+    resamples = jackknife_resampling(data)
+    stats = np.array([statistic(r) for r in resamples])
+    return stats
+
 def jackknife_stat_two_samples(data, data2, statistic, aggregator=None):
     if isinstance(data, np.ndarray): # not block
         jk_X = jackknife_resampling(data)
@@ -288,11 +313,6 @@ def jackknife_stat_two_samples(data, data2, statistic, aggregator=None):
         raise ValueError("data must be a tuple or np.ndarray.")
     return stats
 
-def jackknife_stat_(data, statistic):
-    resamples = jackknife_resampling(data)
-    stats = np.array([statistic(r) for r in resamples])
-    return stats
-
 def _percentile_of_score(a, score, axis, account_equal=False):
     """Vectorized, simplified `scipy.stats.percentileofscore`.
 
@@ -345,7 +365,7 @@ def _resample(data, data2, use_numba, statistic, R, n_min=1, smooth=False, aggre
                 resample_kwargs = dict(aggregator=aggregator)
             else:
                 N = 0
-                sample_stat = np.NaN
+                sample_stat = np.nan
         else:
             if data.ndim == 1:
                 data = data[:, None]
@@ -370,7 +390,7 @@ def _resample(data, data2, use_numba, statistic, R, n_min=1, smooth=False, aggre
             #     data2 = np.hstack(data2)
     return data, data2, theta_hat_b, sample_stat, N
 
-def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=int(1e5), account_equal=False, use_numba=True, n_min=1, aggregator=_nb_mean, **kwargs):
+def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=int(1e5), account_equal=False, use_numba='auto', n_min=1, aggregator=_nb_mean, exclude_nans=False, **kwargs):
     """
     If data2 is provided, assumes a block resampling and statistic takes two arguments.
     Optional kwargs for aggregating data, data2 before computing the statistic:
@@ -378,6 +398,8 @@ def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=i
                          def nb_mean(x):
                              return np.mean(x)
     """
+    if use_numba == 'auto':
+        use_numba = isinstance(statistic, CPUDispatcher)
     if alternative == 'two-sided':
         probs = np.array([alpha/2, 1 - alpha/2])
     elif alternative == 'less':
@@ -390,9 +412,9 @@ def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=i
     data, data2, theta_hat_b, sample_stat, N = _resample(data, data2, use_numba, statistic, R=R, n_min=n_min, aggregator=aggregator, **kwargs)
 
     if theta_hat_b is None:
-        return np.array([np.NaN, np.NaN])
+        return np.array([np.nan, np.nan])
 
-    alpha_bca = _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba, aggregator=aggregator)[0]
+    alpha_bca = _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba, aggregator=aggregator, exclude_nans=exclude_nans)[0]
 
     if np.isnan(alpha_bca).all():
         warnings.warn('CI shows there is only one value. Check data.', RuntimeWarning)
@@ -402,7 +424,7 @@ def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=i
             sample_stat = statistic(data, data2)
         return np.array([sample_stat, sample_stat])
     else:
-        return _compute_CI_percentile(theta_hat_b, alpha_bca, alternative, to_ptg=True)
+        return _compute_CI_percentile(theta_hat_b, alpha_bca, alternative, to_ptg=True, exclude_nans=exclude_nans)
     #elif alternative == 'two-sided':
     #    return  np.percentile(theta_hat_b, alpha_bca*100, axis=0)
     #elif alternative == 'less':
@@ -410,7 +432,7 @@ def CI_bca(data, statistic, data2=None, alternative='two-sided', alpha=0.05, R=i
     #elif alternative == 'greater':
     #    return np.array([np.percentile(theta_hat_b, alpha_bca[0]*100, axis=0), np.inf])
 
-def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba, aggregator=None):
+def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use_numba, aggregator=None, exclude_nans=False):
     """Bias-corrected and accelerated interval."""
     # calculate z0_hat
     if data2 is None:
@@ -421,7 +443,12 @@ def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use
                                   np.array([aggregator(di) for di in data2]))
         else:
             theta_hat = statistic(data, data2)
-    percentile = _percentile_of_score(theta_hat_b, theta_hat, axis=-1, account_equal=account_equal)
+    if exclude_nans:
+        valid = ~(np.isnan(np.atleast_2d(theta_hat_b)).any(axis=1))
+        if not valid.all():
+            print("excluded nans from the resampled statistics")
+            theta_hat_b = theta_hat_b[valid]
+    percentile = _percentile_of_score(theta_hat_b, theta_hat, axis=0, account_equal=account_equal)
     z0_hat = ndtri(percentile)
 
     # calculate a_hat
@@ -430,6 +457,10 @@ def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use
         theta_hat_jk = jackknife_computer(data, statistic)  # jackknife resample
     else:
         theta_hat_jk = jackknife_stat_two_samples(data, data2, statistic, aggregator=aggregator)
+    if exclude_nans:
+        valid = ~(np.isnan(np.atleast_2d(theta_hat_jk)).any(axis=1))
+        theta_hat_jk = theta_hat_jk[valid]
+
     n = theta_hat_jk.shape[0]
     theta_hat_jk_dot = theta_hat_jk.mean(axis=0)
 
@@ -440,10 +471,13 @@ def _bca_interval(data, data2, statistic, probs, theta_hat_b, account_equal, use
 
     # calculate alpha_1, alpha_2
     def compute_alpha(p):
+        z0_hat_expanded = np.atleast_1d(z0_hat)[:, None] # compatibility with multiple outputs
+        a_hat_expanded = np.atleast_1d(a_hat)[: , None]
         z_alpha = ndtri(p)
-        num = z0_hat + z_alpha
-        return ndtr(z0_hat + num/(1 - a_hat*num))
+        num = z0_hat_expanded + z_alpha[None]
+        return ndtr(z0_hat_expanded + num/(1 - a_hat_expanded*num))
     alpha_bca = compute_alpha(probs[(probs != 0) & (probs != 1)])
+    alpha_bca = np.atleast_1d(alpha_bca.squeeze())
     if (alpha_bca > 1).any() or (alpha_bca < 0).any():
         warnings.warn('percentiles must be in [0, 1]. bca percentiles: {}\nForcing percentiles in [0,1]...'.format(alpha_bca), RuntimeWarning)
         alpha_bca = np.clip(alpha_bca, 0, 1)
@@ -545,7 +579,7 @@ def _bootstrap_studentized_resampling(data, statistic, alpha=0.05, R=10000, stud
             return t_result, std_err
     else:
         def get_studentized(data_r, result, seed):
-            return result - base, np.NaN
+            return result - base, np.nan
 
     data_r = resample_nb_X(data, R=R, seed=seed, smooth=smooth)
     if se_func is None:
@@ -578,7 +612,9 @@ def CI_studentized(data, statistic, R=int(1e5), alpha=0.05, alternative='two-sid
         CI = compute_CI_studentized(base, results, studentized_results, alpha=alpha, alternative=alternative)
     return CI
 
-def _compute_CI_percentile(boot_sample, alpha, alternative, to_ptg=False):
+def _compute_CI_percentile(boot_sample, alpha, alternative, to_ptg=False, exclude_nans=False):
+    if isinstance(alpha, np.ndarray) and alpha.ndim == 2: # variable alpha for each output. Used for CI_bca
+        return np.vstack(_compute_CI_percentile(boot_sample[:,i], alpha[i], alternative, to_ptg, exclude_nans) for i in range(alpha.shape[0]))
     alpha_iter = isinstance(alpha, Iterable)
     if alpha_iter:
         alpha = np.asarray(alpha)
@@ -587,23 +623,28 @@ def _compute_CI_percentile(boot_sample, alpha, alternative, to_ptg=False):
         else:
             alpha_ptg = alpha
     else:
-        alpha_ptg = alpha*100 if alpha < 1 else alpha
+        alpha_ptg = alpha*100 if (alpha < 1 or to_ptg) else alpha
     if boot_sample.ndim == 1:
         output_len = 1
     else:
         output_len = boot_sample.shape[1]
+    if exclude_nans:
+        percentile = np.nanpercentile
+    else:
+        percentile = np.percentile
+
     if alternative == 'two-sided':
-        CI = np.percentile(boot_sample, alpha_ptg if alpha_iter else [alpha_ptg/2, 100 - alpha_ptg/2], axis=0).T
+        CI = percentile(boot_sample, alpha_ptg if alpha_iter else [alpha_ptg/2, 100 - alpha_ptg/2], axis=0).T
         CI = np.atleast_2d(CI)
     elif alternative == 'less':
-        CI = np.vstack((-np.inf * np.ones((output_len)), np.percentile(boot_sample, alpha_ptg[0] if alpha_iter else 100-alpha_ptg, axis=0))).T
+        CI = np.vstack((-np.inf * np.ones((output_len)), percentile(boot_sample, alpha_ptg[0] if alpha_iter else 100-alpha_ptg, axis=0))).T
     elif alternative == 'greater':
-        CI = np.vstack((np.percentile(boot_sample, alpha_ptg[0] if alpha_iter else alpha_ptg, axis=0), np.inf * np.ones((output_len)))).T
+        CI = np.vstack((percentile(boot_sample, alpha_ptg[0] if alpha_iter else alpha_ptg, axis=0), np.inf * np.ones((output_len)))).T
     else:
         raise ValueError(f"alternative '{alternative}' not valid. Available: 'two-sided', 'less', 'greater'.")
     return CI
 
-def CI_percentile(data, statistic, data2=None, R=int(1e5), alpha=0.05, smooth=False, alternative='two-sided', n_min=1, use_numba=True, **kwargs):
+def CI_percentile(data, statistic, data2=None, R=int(1e5), alpha=0.05, smooth=False, alternative='two-sided', n_min=1, use_numba='auto', return_resamples=False, **kwargs):
     """
     If data2 is provided, statistic takes two arguments and assumes block resampling if the input data are tuples.
     Optional kwargs for aggregating data, data2 before computing the statistic:
@@ -611,11 +652,17 @@ def CI_percentile(data, statistic, data2=None, R=int(1e5), alpha=0.05, smooth=Fa
                          def nb_mean(x):
                              return np.mean(x)
     """
+    if use_numba == 'auto':
+        use_numba = isinstance(statistic, CPUDispatcher)
     data, data2, boot_sample, sample_stat, N = _resample(data, data2, use_numba, statistic, R=R, n_min=n_min, smooth=smooth, **kwargs)
     if boot_sample is None:
-        return np.array([np.NaN, np.NaN])
+        CI = np.array([np.nan, np.nan])
     else:
-        return _compute_CI_percentile(boot_sample, alpha, alternative)
+        CI = _compute_CI_percentile(boot_sample, alpha, alternative)
+    if return_resamples:
+        return CI, boot_sample
+    else:
+        return CI
 
 def CI_all(data, statistic, R=int(1e5), alpha=0.05, alternative='two-sided', coverage_iters=int(1e5), coverage_seed=42, avg_len=3, exclude=['studentized_vs', 'studentized_vs_smooth']):
     """
