@@ -43,7 +43,7 @@ def CI_specs(CIs, data, stat, coverage_iters=int(1e4), seed=42, avg_len=3, **cov
     CIs2 = CIs.copy()
     CIs2['width'] = np.hstack([np.diff(CI) for CI in CI_arr])
     CIs2['asymmetry'] = np.hstack([(CI[1] - sample_stat) / (sample_stat - CI[0]) for CI in CI_arr])
-    env = locals()    
+    env = locals()
     for end in ['avg', 'last']:
         for k in ['low_fails', 'high_fails', 'coverage']:
             key = f"{k}_{end}"
@@ -62,7 +62,7 @@ def find_best(CIs, data=None, stat=None, alpha=0.05, alternative=None, alpha_mar
             alternative = 'greater'
         else:
             alternative = 'two-sided'
-    
+
     coverages_last, coverages_avg = CIs[['coverage-last', 'coverage-avg']].values.T
     valid = np.unique(np.hstack([np.where(coverages_last >= (1-alpha_expanded_last))[0],
                                  np.where(coverages_avg >= (1-alpha_expanded_avg))[0],
@@ -75,39 +75,46 @@ def find_best(CIs, data=None, stat=None, alpha=0.05, alternative=None, alpha_mar
         best_interval = CIs['low'].values[valid].argmax()
     return CIs.iloc[valid].iloc[best_interval]
 
-def ci_percentile_equal_tailed(x, p, alpha=0.05, alternative='two-sided'):
+def ci_percentile_equal_tailed(x, p, alpha=0.05, alternative='two-sided', x_H1=None):
     """
-    Exact confidence interval for percentiles. 
+    Exact confidence interval for percentiles.
     Attrs:
+        - x: data
         - p:  percentile in (0, 1)
-            
+        - alpha: significance level
+        - alternative: 'two-sided', 'less', 'greater'
+        - x_H1: if provided, returns the p_value:
+            - 'two-sided': 2 * min(p_value_less, p_value_greater). H1: x != x_H1
+            - 'less': upper bound. H1: x < x_H1
+            - 'greater': lower bound. H1: x > x_H1
+
     Returns CI [Yi, Yj] such that Prob(x in CI) => 1 - alpha. Yi, Yj are order statistics.
     https://online.stat.psu.edu/stat415/book/export/html/835
     """
     n = x.size
     if n <= 2:
         warnings.warn(f"n = {n} is too small. Returning NaN", RuntimeWarning)
-        return np.NaN
+        return np.nan
     else:
-        if n < 500:    
-            def binom_cdf(x): 
+        if n < 500:
+            def binom_cdf(x):
                 return np.array([math.comb(n, k) * p**k * (1-p)**(n-k) for k in range(x+1)]).sum()
             binom_cdf = np.vectorize(binom_cdf)
-            #def binom_pmf(x):    
+            #def binom_pmf(x):
             #    return math.comb(n, x) * p**x * (1-p)**(n-x)
         else:
             try:
                 from scipy.stats import binom
                 binom_cdf = lambda x: binom.cdf(x, n, p)
             except:
-                raise ImportError('scipy not found. Please install it for n > 500. Alternatively, use the normal aproximation for the binomial.')       
-        
+                raise ImportError('scipy not found. Please install it for n > 500. Alternatively, use the normal aproximation for the binomial.')
+
         p_below_percentile = binom_cdf(np.arange(n+1))
-        
+
         if alternative == 'two-sided':
             lows = np.where(p_below_percentile <= alpha/2)[0]
             if lows.size > 0:
-                l = lows[-1] + 1 
+                l = lows[-1] + 1
             else:
                 warnings.warn('n is too small to warrantee an exact lower bound other than the minimum.', RuntimeWarning)
                 l = 0
@@ -130,7 +137,7 @@ def ci_percentile_equal_tailed(x, p, alpha=0.05, alternative='two-sided'):
             else:
                 q1 = p_below_percentile[u+1]
                 CI_prob1 = (u+1) / n # add 1 to both endpoints (indexing in python starts at 0)
-            quantiles = np.array([q0, q1]) 
+            quantiles = np.array([q0, q1])
             CI_prob = np.array([CI_prob0, CI_prob1])
             CI = np.sort(x)[ci_equal_tailed]
         elif alternative == 'less':
@@ -153,13 +160,27 @@ def ci_percentile_equal_tailed(x, p, alpha=0.05, alternative='two-sided'):
             quantiles = p_below_percentile[l-1]
             CI = np.array([np.sort(x)[l], np.inf])
             CI_prob = l / n
-        return CI, CI_prob, quantiles
+        if x_H1 is not None:
+            observed_k_less = (x < x_H1).sum()
+            p_value_less = 1 - p_below_percentile[observed_k_less] # probability of obtaining at least k successes assuming the null hypothesis is true
+
+            observed_k_greater = (x > x_H1).sum()
+            p_value_greater = 1 - p_below_percentile[observed_k_greater] # probability of obtaining at least k successes assuming the null hypothesis is true
+            if alternative == 'less':
+                p_value = p_value_less
+            elif alternative == 'greater':
+                p_value = p_value_greater
+            else:
+                p_value = 2 * min(p_value_less, p_value_greater)
+            return CI, p_value, CI_prob, quantiles
+        else:
+            return CI, CI_prob, quantiles
 
 
 def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.005):
     """
     DO NOT USE. This does not guarantee equal tailed. It is here as a reminder.
-    
+
     Confidence 1 - alpha that the interval (Yi, Yj) contains the percentile p. Yi, Yj are order statistics.
     Attrs:
         - p:  percentile in (0, 1)
@@ -167,7 +188,7 @@ def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.00
     """
     x_sorted = np.sort(x)
     n = x.size
-    def confidence(low, high):    
+    def confidence(low, high):
         return np.array([math.comb(n, k) * p**k * (1-p)**(n-k) for k in range(low, high)]).sum()
     c = {}
     if alternative == 'two-sided':
@@ -178,7 +199,7 @@ def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.00
         for high in range(1, n+1):
             c[high-1] = confidence(0, high)
     elif alternative == 'greater':
-        for low in range(1, n+1): 
+        for low in range(1, n+1):
             c[low-1] = 1 - confidence(0, low)
     c = pd.Series(c)
     c_pruned = c[(c - (1-alpha)).abs() < d_alpha]
@@ -194,7 +215,7 @@ def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.00
     else:
         bound = [x_sorted[i] for i in c_pruned.index]
         if alternative == 'greater':
-            c_pruned['low'] = bound 
+            c_pruned['low'] = bound
             c_pruned['high'] = np.inf
         else:
             c_pruned['low'] = -1 * np.inf
@@ -204,7 +225,7 @@ def _exact_ci_percentile(x, p, alpha=0.05, alternative='two-sided', d_alpha=0.00
 def _find_ci_percentile_exact_confidence(ci, x, p, tol=1e-3):
     """
     DO NOT USE. This does not guarantee equal tailed CI.
-    
+
     Given a CI 'ci' from data 'x' of a percentile statistic, return the exact confidence it provides.
     """
     if not math.isfinite(ci[0]):
@@ -217,4 +238,4 @@ def _find_ci_percentile_exact_confidence(ci, x, p, tol=1e-3):
         alternative = 'two-sided'
         prune = lambda c: c[((c['low'] - ci[0]).abs() < tol) & ((c['high'] - ci[1]).abs() < tol)]
     c = _exact_ci_percentile(x, p, alternative=alternative, d_alpha=0.5)
-    return prune(c)
\ No newline at end of file
+    return prune(c)
